"""Photovoltaic farm model definition."""
import pandas as pd
import xarray as xr
from ..utils.climate import get_temperature_at_height
from ..utils.radiation import HorizontalRadiationComputer
from ..utils.radiation import TiltedRadiationComputer
from ..data_source import Composer
from ..actuator_base import ExtractorBase

#: Available-dataset names.
VARIABLE_NAMES = [
    'global_horizontal_et', 'global_horizontal_surf',
    'global_tilted_surf', 'generation', 'capacity_factor',
    'cell_efficiency'
]


class Actuator(ExtractorBase):
    """Photovoltaic model."""
    #: Default result name.
    DEFAULT_RESULT_NAME = 'capacity_factor'

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Naming constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Actuator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        """
        if result_mng.name != self.DEFAULT_RESULT_NAME:
            self.log.warning(
                'Result name {} given to constructor does not correspond '
                'to {} to be estimated by {}'.format(
                    result_mng.result_name, self.DEFAULT_RESULT_NAME,
                    self.name))

        super(Actuator, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg,
            variable_names=VARIABLE_NAMES, **kwargs)

        #: Required input-variables.
        self.input_variable_names = {'surface_downward_radiation',
                                'surface_temperature'}

        #: Global horizontal surface irradiance.
        self.global_horizontal_surf = None
        #: Global horizontal extraterrestrial irradiance.
        self.global_horizontal_et = None
        #: Global tilted surface irradiance.
        self.global_tilted_surf = None
        #: Clearness index.
        self.clearness_index = None
        #: Generation.
        self.generation = None
        #: Cell efficiency.
        self.cell_efficiency = None
        #: Capacity factor.
        self.capacity_factor = None

    def transform(self, data_src, stage=None, **kwargs):
        """Compute photovoltaic electricity generation from climate data.

        :param data_src: Input data source.
        :param stage: Modeling stage: `'fit'` or `'predict'`.
          May be required if features differ in prediction stage.
        :type data_src: :py:class:`..grid.GriddedDataSourceBase`
        :type stage: str

        :returns: Transformed dataset.
        :rtype: :py:class:`xarray.Dataset`
        """
        # Add regions domain cropping
        functions = [data_src.crop_area]

        if hasattr(self.result_mng, 'modifier'):
            if (hasattr(self.result_mng.modifier.cfg, 'stage') and
                    (self.result_mng.modifier.cfg['stage'] != stage)):
                pass
            else:
                # Add modifier transformation
                functions.append(self.result_mng.modifier.transform)

        # Add get generation and get regional mean
        functions.extend([self.get_generation, data_src.get_regional_mean])

        # Get pv generation from climate data
        ds = data_src.load(transform=Composer(*functions),
                           variable_names=self.input_variable_names)

        return ds

    def get_extractor_postfix(self, **kwargs):
        """Get extractor postfix.

        returns: Postfix.
        rtype: str
        """
        postfix = '{}_{}'.format(
            super(Actuator, self).get_extractor_postfix(**kwargs),
            self.med.cfg['frequency'])

        return postfix

    def get_generation(self, ds, data_src, **kwargs):
        """Compute the photovoltaic electricity generation
        from the daily climate dataset.

        :param ds: Dataset containing climate variables from which to
          compute generation.
        :param data_src: Data source object from which to get features.
        :type ds: :py:class:`xarray.Dataset`
        :type data_src: :py:class:`..grid.GriddedDataSourceBase`

        :returns: A dataset containing the electricity generation.
        :rtype: :py:class:`xarray.Dataset`

        .. note::
          * The dataset should contain a `lat` and a `lon` coordinate
            variables.
          * The dataset should contain a `surface_downward_radiation` variable
            for the global horizontal radiation at the surface.
          * If the dataset contains a `surface_temperature` variable
            for the surface temperature,
            the temperature dependent cell efficiency is computed.
          * If the dataset contains a `surface_temperature` variable
            for the surface temperature,
            the temperature dependent cell efficiency is computed.

        .. warning:: Time index in dataset :py:obj:`ds` should be UTC.
        """
        # Get HOURLY time index (UTC)
        time = ds.indexes['time']
        freq_data = time.inferred_freq.upper()
        if freq_data in ['D', '1D']:
            time = pd.date_range(start=time[0], freq='H',
                                 end=(time[-1] + pd.Timedelta(23, 'H')))

        # Get surface temperature
        temp_a = None
        if 'surface_temperature' in ds:
            temp_0 = ds['surface_temperature']
            z_0 = float(temp_0.attrs[data_src.cfg['height']])

            # Adjust to surface temperature (2m)
            temp_a = get_temperature_at_height(temp_0, z_0, 2.)

        # Get hourly global horizontal extraterrestrial radiation
        if not self.cfg.get('no_verbose'):
            self.log.info(
                'Getting hourly global extraterrestrial solar radiation')
        rc_hour = HorizontalRadiationComputer(
            time, ds.lat, ds.lon,
            global_horizontal_surf=ds['surface_downward_radiation'],
            angles_in_degrees=True)
        self.global_horizontal_et = rc_hour.global_horizontal_et

        # Recycle solar computer
        sc_hour = rc_hour.sc

        if freq_data in ['D', '1D']:
            # Up-sample temperature
            temp_a = temp_a.resample(time='H').ffill().reindex(
                time=time, method='ffill')

            # Get daily-mean global horizontal extraterrestrial radiation
            if not self.cfg.get('no_verbose'):
                self.log.info(
                    'Getting daily-mean global horizontal extraterrestrial '
                    'solar radiation')
            global_horizontal_et_day = self.global_horizontal_et.resample(
                time='D').mean('time', keep_attrs=True)
            global_horizontal_surf_day = ds['surface_downward_radiation']

            # Get clearness index from DAILY-averaged
            # toa and surface radiations
            if not self.cfg.get('no_verbose'):
                self.log.info('Getting daily-mean clearness index')
            rc_day = HorizontalRadiationComputer(
                time, ds.lat, ds.lon,
                global_horizontal_et=global_horizontal_et_day,
                global_horizontal_surf=global_horizontal_surf_day,
                angles_in_degrees=True)
            clearness_index_day = rc_day.clearness_index

            # Up-sample daily clearness index to hourly
            if not self.cfg.get('no_verbose'):
                self.log.info('Up-sampling clearness index to hours')
            self.clearness_index = clearness_index_day.resample(
                time='H').ffill().reindex(time=time, method='ffill')

            # Radiations computation from clearness index and hourly
            # global horizontal extraterrestrial radiation
            if not self.cfg.get('no_verbose'):
                self.log.info(
                    'Getting hourly global horizontal surface radiation from '
                    'hourly horizontal extraterrestrial radiation and '
                    'up-sampled clearness index')
            rc_hour = TiltedRadiationComputer(
                time, ds.lat, ds.lon,
                global_horizontal_et=self.global_horizontal_et,
                clearness_index=self.clearness_index,
                angles_in_degrees=True, **self.cfg['surface'],
                solar_computer=sc_hour)

            self.global_horizontal_surf = rc_hour.global_horizontal_surf

        elif freq_data in ['H', '1H']:
            # Radiations computation from hourly global horizontal
            # extraterrestrial and surface radiation (from source)
            self.global_horizontal_surf = ds['surface_downward_radiation']
            rc_hour = TiltedRadiationComputer(
                time, ds.lat, ds.lon,
                global_horizontal_et=self.global_horizontal_et,
                global_horizontal_surf=self.global_horizontal_surf,
                angles_in_degrees=True, **self.cfg['surface'],
                solar_computer=sc_hour)

        if not self.cfg.get('no_verbose'):
            self.log.info(
                'Computing hourly global tilted surface radiation')
        self.global_tilted_surf = rc_hour.global_tilted_surf

        # Get hourly solar generation
        if not self.cfg.get('no_verbose'):
            self.log.info('Computing hourly photovoltaic generation')
        self.generation, self.cell_efficiency = get_pv_power(
            self.global_tilted_surf, temp_a, **self.cfg['module'],
            return_efficiency=True)

        # Get hourly capacity factor
        self.capacity_factor = (
            self.generation / self.cfg['module']['nominal_power'])

        # Save key variables in object (e.g. for plots)
        if self.med.cfg['frequency'] == 'day':
            # Add daily-integrated variables to dataset
            if not self.cfg.get('no_verbose'):
                self.log.info('Resampling from hourly to daily frequency')
            for variable_name in VARIABLE_NAMES:
                setattr(self, variable_name, getattr(
                    self, variable_name).resample(time='D').sum(
                        'time', keep_attrs=True))

            # Manage daily units
            generation_units = 'Wh/d'
            self.capacity_factor /= 24
        else:
            generation_units = 'Wh'

        try:
            self.capacity_factor.name = self.result_mng.result_name
            self.capacity_factor.attrs['long_name'] = (
                'Photovoltaic Capacity Factor')
            self.capacity_factor.attrs['units'] = ''
            self.generation.attrs['units'] = generation_units
        except AttributeError:
            pass

        ds = xr.Dataset({variable_name: getattr(self, variable_name)
                         for variable_name in VARIABLE_NAMES})

        return ds


def get_pv_power(global_tilted_surf, temp_a=None, wind_speed=1., area=1.675,
                 n_modules=1, eff_elec=0.86, eff_cell=None, thermal_loss=0.004,
                 temp_ref=298.15, temp_cell_noct=319.15, nominal_power=None,
                 return_efficiency=False):
    """Set photovoltaic power from inclinaison, clearness index
    and surface air temperature.

    :param global_tilted_surf: Global tilted surface radiation
      (W/m2, or Wh/m2/h).
    :param temp_a: Surface air temperature (K).
      If None, the temperature of the cell is considered to be the same
      as the reference temperature temp_ref.
      Default is None.
    :param wind_speed: Magnitude of the wind velocity (m/s).
    :param area: Area of each module (m2). Default is 1.675 (m2).
    :param n_modules: Number of modules. Default is 1.
    :param eff_elec: Efficiency of the electronics
      (inverter, array and power conditionning, etc.). Default is 0.86.
    :param eff_cell: Efficiency of the cell at the reference temperature
      temp_ref. If None, compute it from the nominal power. Default is None.
    :param thermal_loss: Efficiency loss per degree Kelvin
      away from the reference temperature. Default is
      :math:`0.0041~\mathrm{K}^{-1}`.
    :param temp_ref: Reference temperature (K). Default 298.15 K.
    :param temp_cell_noct: Cell temperature at the
      Nominal Operating Cell Temperature (NOCT) (K).
      Default is 319.15 K.
    :param nominal_power: Nominal power of the module (W). See note.
      Only used to compute the efficiency of the cell eff_cell
      at the reference temperature temp_ref if eff_cell is not given.
    :param return_efficiency: If True, return the temperature dependent
      cell efficiency. Default is False.
    :type global_tilted_surf: (sequence of) :py:class:`float`
    :type temp_a: (sequence of) :py:class:`float`
    :type wind_speed: (sequence of) :py:class:`float`
    :type area: float
    :type n_modules: int
    :type eff_elec: float
    :type eff_cell: float
    :type thermal_loss: float
    :type temp_ref: float
    :type temp_cell_noct: float
    :type nominal_power: float
    :type return_efficiency: bool

    :returns: Electric power delivered by the array (W/m2, or Wh/m2/h).
    :rtype: (:py:class:`tuple` of) (sequence of) :py:class:`float`

    .. warning::
      The global tilted radiation `global_tilted_surf` should be given as
      power in W/m2, or as averaged power in Wh/m2/h to avoid erros
      in cell efficiency computations.
    """
    # Temperature dependent efficiency of the cell
    eff_cell_t = get_cell_efficiency(
        temp_a=temp_a, global_tilted_surf=global_tilted_surf, area=area,
        wind_speed=wind_speed, eff_cell=eff_cell, thermal_loss=thermal_loss,
        temp_ref=temp_ref, temp_cell_noct=temp_cell_noct,
        nominal_power=nominal_power)

    # Electric power delivered by the array
    generation = area * n_modules * eff_elec * eff_cell_t * global_tilted_surf

    # Name power if possible
    try:
        generation.name = 'generation'
        generation.attrs['long_name'] = 'Photovoltaic Array Generation'
        generation.attrs['units'] = (
            '{:s}'.format(global_tilted_surf.units.strip('/ m2'))
            if hasattr(global_tilted_surf, 'units') else 'Wh')
    except AttributeError:
        pass

    return (generation, eff_cell_t) if return_efficiency else generation


def get_cell_efficiency(
        temp_a=None, global_tilted_surf=800., wind_speed=1., area=1.,
        eff_cell=None, thermal_loss=0.004, temp_ref=298.15,
        temp_cell_noct=319.15, nominal_power=None):
    """Get temperature dependent cell efficiency.

    :param temp_a: Surface air temperature (K).
      If None, the temperature of the cell is considered to be the same
      as the reference temperature temp_ref.
      Default is None.
    :param global_tilted_surf: Global tilted radiation (W/m2, or Wh/m2/h).
      Default is the :py:obj:`global_tilted_surf` at
      Nominal Operating Cell Temperature (_noct).
    :param wind_speed: Magnitude of the wind velocity (m/s).
    :param area: Area of each module (m2). Default is 1 (m2).
      (inverter, array and power conditionning, etc.). Default is 0.86.
    :param eff_cell: Efficiency of the cell at the reference temperature
      temp_ref. If None, compute it from the nominal power. Default is None.
    :param thermal_loss: Efficiency loss per degree Kelvin
      away from the reference temperature. Default is
      :math:`0.0041~\mathrm{K}^{-1}`.
    :param temp_ref: Reference temperature (K). Default 298.15 K.
    :param temp_cell_noct: Cell temperature at _noct (K). Default is 319.15 K.
    :param nominal_power: Nominal power of the module (W). See note.
      Only used to compute the efficiency of the cell eff_cell
      at the reference temperature temp_ref if eff_cell is not given.
    :type temp_a: (sequence of) :py:class:`float`
    :type global_tilted_surf: (sequence of) :py:class:`float`
    :type wind_speed: (sequence of) :py:class:`float`
    :type area: float
    :type eff_cell: float
    :type thermal_loss: float
    :type temp_ref: float
    :type temp_cell_noct: float
    :type nominal_power: float

    :returns: Temperature dependent cell efficiency.
    :rtype: (sequence of) float)

    .. note::
        * The thermal model of Duffie and Beckman (2013)
          is used here to compute the efficiency of the cell
          for a given ambient temperature.
        * The global tilted radiation :py:obj:`global_tilted_surf` should be
          given as power in W/m2, or as averaged power in Wh/m2/h to avoid
          erros in cell efficiency computations.

    .. seealso::
        Duffie, J.A., Beckman, W.A., 2013.
        *Solar Energy and Thermal Processes*, fourth ed. Wiley, Hoboken, NJ.
    """
    # Radiation on module plane at _noct (W/m2)
    global_tilted_surf_noct = 800.
    # Ambient Temperature at _noct (K)
    temp_a_noct = 293.15
    # Light intensity at STC (W/m2). See note.
    ir_std = 1000.

    # If not given, compute cell efficiency from nominal power
    if eff_cell is None:
        eff_cell = (nominal_power / ir_std) / area

    if temp_a is None:
        # If the ambient temperature is not given, the cell temperature
        # is taken as the same as the reference temperature. As a result,
        # the cell efficiency does not depend on the temperature.
        temp_cell = temp_ref
    else:
        # Compute temperature dependent efficiency of module
        # Thermal loss ration: :math:`U_{L, _noct} / U_L`
        loss_ratio = 9.5 / (5.7 + 3.8 * wind_speed)

        # Cell temperature (K). See note.
        # :math:`  au \alpha / U_L = (T_{c, _noct} - T_{a, _noct}) / G_{T, NOCT}`
        temp_cell = (temp_a + (global_tilted_surf / global_tilted_surf_noct)
                     * loss_ratio * (temp_cell_noct - temp_a_noct))

    # Temperature dependent efficiency of the cell
    eff_cell_t = eff_cell * (1 - thermal_loss * (temp_cell - temp_ref))

    # Name
    try:
        eff_cell_t.name = 'cell_efficiency'
        eff_cell_t.attrs['long_name'] = 'Wheather Dependent Cell Efficiency'
        eff_cell_t.attrs['units'] = ''
    except AttributeError:
        pass

    return eff_cell_t



"""GISCO API for European borders."""
import requests
from ..geo import DefaultMaskAPI, DefaultMaskMaker


class DataSource(DefaultMaskMaker, DefaultMaskAPI):
    """GISCO data source."""
    #: Default source name.
    DEFAULT_SRC_NAME = 'gisco'

    #: Flag preventing several downloads per area.
    ONE_DOWNLOAD_PER_AREA = True

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor calling :py:class:`DefaultMaskMaker`'s
        constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    def get_url(self, **kwargs):
        """ Get URL of geographical data.

        :returns: URL pointing to the geographical data.
        :rtype: str
        """
        # Get the list of files
        path = self.cfg['host'] + '/' + self.cfg['path'] + '/'
        url_files = path + 'nuts-{}-files.json'.format(self.cfg['year'])
        n_trials = 0
        while n_trials < self.cfg['max_fetch_trials']:
            try:
                # Request and raise exception if needed
                response = requests.get(url_files)
                response.raise_for_status()
                break
            except requests.exceptions.SSLError as e:
                # Retry
                self.log.warning(
                    '{} trial to fetch {} file list failed: {}'.format(
                        n_trials + 1, self.namestr(e)))
                n_trials += 1
                continue
        # Verify that last trial succeeded
        if n_trials >= self.cfg['max_fetch_trials']:
            # All trials failed
            self.log.critical(
                'Fetching failed after {:d} trials'.format(n_trials))
            raise RuntimeError

        # Get the filename
        filename = self.get_filename(**kwargs)

        # Get URL
        url = '{}{}'.format(path, response.json()['geojson'][filename])

        return url

    def get_filename(self, *args, **kwargs):
        """Get downloaded-file name.

        :returns: Filename.
        :rtype: str
        """
        fmt = (self.cfg['spatial_type'], self.cfg['scale'],
               self.cfg['year'], self.cfg['projection'],
               self.cfg['nuts_level'])

        return 'NUTS_{}_{}_{}_{}_LEVL_{}.geojson'.format(*fmt)



"""RTE-bilan API."""
from pkg_resources import resource_stream
import pandas as pd
import xarray as xr
from ..data_source import DataSourceLoaderBase
from ..component import (DataSourceWithComponentsMixin,
                         parse_variable_component_args,
                         finalize_dataset)


class DataSource(DataSourceLoaderBase, DataSourceWithComponentsMixin):
    #: Default source name.
    DEFAULT_SRC_NAME = 'rte_bilan'

    #: Area: `'France'`.
    AREA = 'France'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    def download(self, *args, **kwargs):
        """Convenience function to warn that RTE-bilan data is not downloaded
        but loaded from package data.

        :returns: Names of variables to download.
        :rtype: :py:class:`set` of :py:class:`str`
        """
        self.log.warning(
            '{} is not to be downloaded. It is instead directly loaded from '
            'the package data.'.format(self.DEFAULT_SRC_NAME))

        return {'capacity_factor'}

    @parse_variable_component_args
    @finalize_dataset
    def load(self, variable_component_names, **kwargs):
        """Load data.

        :param variable_component_names: Names of components to load per
          variable. Default is `None`, in which case all components in
          :py:attr::`med.component_managers` are loaded.
        :type variable_component_names: mapping from :py:class:`str`
          to collection

        :returns: Time series for each variable and component.
        :rtype: dict
        """
        # Get places and regions
        src_place_regions, _ = self.med.geo_src.get_place_regions_for_source(
            self.name, self.AREA, **kwargs)

        ds = {}
        for variable_name, component_names in variable_component_names.items():
            # Manage wind component as onshore only
            component_name = 'wind'
            component_to_load_names = component_names.copy()
            if component_name in component_names:
                # Make sure that generation and capacity are loaded
                component_to_load_names.add('wind-onshore')

                # Remove capacity factor from components to load
                component_to_load_names.discard(component_name)

            da = None
            for component_name in component_to_load_names:
                # Read preprocessed RTE report for this variable and component
                resource_name = self.get_download_resource_name(
                    variable_name, component_name, **kwargs)
                with resource_stream(__name__, resource_name) as f:
                    df = pd.read_csv(f, index_col=0)

                # Define time index
                time = pd.to_datetime(
                    {'year': df.index.values.astype(str),
                     'month': 12, 'day': 31})
                df.index = pd.DatetimeIndex(time, freq='A-DEC')
                df.index.name = 'time'

                # Group by zones, averaging capacity factors
                df_zones = pd.DataFrame(
                    0., index=df.index, columns=list(src_place_regions))
                for place_name, region_names in src_place_regions.items():
                    df_zones[place_name] = df[region_names].mean('columns')

                # Create array and add component coordinate
                da_comp = xr.DataArray(
                    df_zones, dims=['time', 'region']).expand_dims(
                    'component').assign_coords(component=[component_name])

                # Add component data to array
                da = da_comp if da is None else xr.concat(
                    [da, da_comp], dim='component')

            # Add total wind component as onshore wind if needed
            component_name = 'wind'
            if component_name in component_names:
                coord_component = da.indexes['component'].tolist()
                idx = coord_component.index('wind-onshore')
                coord_component[idx] = component_name
                da['component'] = coord_component

            # Add variable to dataset
            ds[variable_name] = da

        return ds

    def get_download_resource_name(self, variable_name, component_name,
                                   **kwargs):
        """Get package resource name for data copied from RTE reports.

        :param variable_name: Variable name.
        :param component_name: Component name.
        :type variable_name: str
        :type component_name: str

        :returns: Resource name.
        :rtype: str
        """
        filename = '{}_{}_year_{}_{}.csv'.format(
            variable_name, component_name, self.cfg['first_year'],
            self.cfg['last_year'])
        resource_name = '../data/{}/{}/{}'.format(
            self.AREA, self.DEFAULT_SRC_NAME, filename)

        return resource_name



"""Extractor defining transform method from functions composition."""
from importlib import import_module
from collections import OrderedDict
from ..data_source import Composer
from ..actuator_base import ExtractorBase
from e4clim.container import ensure_collection


class Actuator(ExtractorBase):
    """Extractor defining transform method from functions composition."""

    def __init__(self, result_mng, name, cfg=None, functions=None,
                 **kwargs):
        """Constructor defining transform method from functions composition.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Actuator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping

        .. note:: Functions to compose must be provided as a list
          from the `functions` entry of the configuration.
        """
        super(Actuator, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg, **kwargs)

        #: Functions composed to define :py:meth:`transform` method.
        self.functions = None

        if functions is None:
            # Import functions
            function_paths = ensure_collection(self.cfg['functions'])
            self._functions = [import_function(function_path)
                               for function_path in function_paths]
        else:
            # Assign user-defined functions
            self._functions = functions

        # Define :py:meth:`transform` method by composing functions
        self.transform = Composer(*self._functions)

        # Add data sources
        if 'data' in self.cfg:
            if self.data_sources is None:
                self.data_sources = OrderedDict()
            self.add_data_sources(**kwargs)

    def transform(self, **kwargs):
        """Empty concrete definition of transform method replaced
        by composed functions during construction."""
        pass


def import_function(function_path):
    """Import function.

    :param function_path: Function path.
    :type function_path: str

    :return: Imported function.
    :rtype: function
    """
    # Get module path and function name
    fun_path_split = function_path.split('.')
    module_path = '.'.join(fun_path_split[:-1])
    function_name = fun_path_split[-1]

    # Import function module
    module = import_module(module_path)

    # Return function
    return getattr(module, function_name)



"""Copernicus Land Cover API."""
import os
import numpy as np
import pandas as pd
import xarray as xr
from fiona.crs import from_epsg
import rasterio
from shapely.geometry import Point
import geopandas as gpd
from pkg_resources import resource_stream
from ..geo import get_mask_path
from ..data_source import DataSourceLoaderBase


class DataSource(DataSourceLoaderBase):
    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or 'clc'
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    def download(self, *args, **kwargs):
        """Warn that :py:meth:`download` method not implemented."""
        self.log.warning(
            '{} data download not implemented'.format(self.name))

    def load(self, *args, **kwargs):
        # Get configuration file to use and import it
        data_dir = self.cfg['dir'] if 'dir' in self.cfg \
            else self.med.cfg['data_dir']
        crs_wgs84 = from_epsg(4326)

        # Read the mask
        mask_filepath = get_mask_path(self.cfg, self.name)
        ds_mask = xr.load_dataset(mask_filepath)

        # Get regular grid limits
        lat, lon = ds_mask.lat.values, ds_mask.lon.values
        if len(ds_mask.lat.shape) > 1:
            n_lat, n_lon = ds_mask.lat.shape
        else:
            n_lat, n_lon = len(ds_mask.lat), len(ds_mask.lon)
            lat, lon = np.tile(lat, (n_lon, 1)).T, np.tile(lon, (n_lat, 1))
        dlat, dlon = np.empty((n_lat + 1, n_lon)), np.empty((n_lat, n_lon + 1))
        dlat[1:-1] = (lat[1:] - lat[:-1]) / 2
        dlat[0] = dlat[1]
        dlat[-1] = dlat[-2]
        dlon[:, 1:-1] = (lon[:, 1:] - lon[:, :-1]) / 2
        dlon[:, 0] = dlon[:, 1]
        dlon[:, -1] = dlon[:, -2]
        lat_lim = np.empty((n_lat + 1, n_lon))
        lat_lim[:-1] = lat - dlat[:-1]
        lat_lim[-1] = lat[-1] + dlat[-1]
        lon_lim = np.empty((n_lat, n_lon + 1))
        lon_lim[:, :-1] = lon - dlon[:, :-1]
        lon_lim[:, -1] = lon[:, -1] + dlon[:, -1]
        # Back to 1D (other grids not yet implemented)
        lat_lim, lon_lim = lat_lim[:, 0], lon_lim[0]

        # Read land cover - roughness length map (CLC/Global Wind Atlas)
        resource_name = '../data/' + 'landuse_roughness_GWA1.csv'
        with resource_stream(__name__, resource_name) as f:
            if not self.cfg.get('no_verbose'):
                self.log.info(
                    'Reading land use to roughness map from {}'.format(
                        resource_name))
            df_map = pd.read_csv(f, index_col=0, header=0).loc[
                :, 'Roughness_Length']
        d_map = df_map.dropna().to_dict()

        # Read the raster file
        filename = 'g' + self.cfg['resolution'] + '_' + 'clc' \
            + self.cfg['year'][2:] + '_V' + self.cfg['version'] \
            + '.' + self.cfg['fileFormat']
        filepath = os.path.join(data_dir, filename)
        if not self.cfg.get('no_verbose'):
            self.log.info('Reading land cover from {}'.format(filepath))
        ds_land = rasterio.open(filepath, 'r')

        # Get window
        lat0, lon0 = lat.min(), lon.min()
        pt0 = gpd.GeoSeries(Point(lon0, lat0),
                            crs=crs_wgs84).to_crs(ds_land.crs)
        latf, lonf = lat.max(), lon.max()
        ptf = gpd.GeoSeries(Point(lonf, latf),
                            crs=crs_wgs84).to_crs(ds_land.crs)
        xmin, ymin, xmax, ymax = float(pt0.x), float(
            pt0.y), float(ptf.x), float(ptf.y)
        window = ds_land.window(xmin, ymin, xmax, ymax)
        window = window.round_offsets().round_lengths()

        # Values
        values = ds_land.read(window=window)[0]

        jleft, itop, jright, ibot = window.flatten()
        jright += jleft
        ibot += itop
        count = xr.zeros_like(ds_mask.mask)
        roughness = xr.zeros_like(ds_mask.mask).astype(float)
        roughness.name = 'roughness_length'
        si, sj = self.cfg['sampling'], self.cfg['sampling']
        if not self.cfg.get('no_verbose'):
            self.log.info(
                'Averaging the land cover over the {} grid'.format(self.name))
        for i in range(itop, ibot, si):
            for j in range(jleft, jright, sj):
                # Get the roughness length and continue only if
                # it has a meaning
                rl = d_map.get(values[i - itop, j - jleft])
                if rl is not None:
                    # Get the longitude and latitude of this point
                    xy = ds_land.xy(i, j)
                    pt = gpd.GeoSeries(
                        Point(*xy), crs=ds_land.crs).to_crs(crs_wgs84)
                    lon_pt, lat_pt = pt[0].coords[0]
                    # Find grid box of the mask to which the land cover point
                    # belongs
                    ilat_mask = np.searchsorted(
                        lat_lim, lat_pt, side='right') - 1
                    if (ilat_mask >= 0) & (ilat_mask < n_lat):
                        jlon_mask = np.searchsorted(
                            lon_lim, lon_pt, side='right') - 1
                        if (jlon_mask >= 0) & (jlon_mask < n_lon):
                            roughness[ilat_mask, jlon_mask] += rl
                            count[ilat_mask, jlon_mask] += 1
        roughness /= count

        # Save the dataset
        postfix = '_{}{}_{}'.format(
            'clc', str(self.cfg['year']), self.name)
        filename = 'roughness_length' + postfix + '.nc'
        filepath = os.path.join(self.med.cfg['data_dir'], 'climate', filename)
        if not self.cfg.get('no_verbose'):
            self.log.info('Saving roughness length to '.format(filepath))
        roughness.to_netcdf(filepath)

        ds_land.close()
        ds_mask.close()



"""Geographic-data and mask base definitions."""
import os
from collections import OrderedDict
from abc import ABC, abstractmethod
from pkg_resources import resource_stream
import requests
import zipfile
from io import BytesIO
from shapely.geometry import Point
from fiona.crs import from_epsg
import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
import oyaml as yaml
import difflib
from .data_source import DataSourceBase, DataSourceLoaderBase, MultiDataSource
from .container import ensure_collection

#: Geographic-variable name.
GEO_VARIABLE_NAME = 'border'


class GeoDataSourceBase(DataSourceBase):
    """Geographic data-source abstract base class."""

    def __init__(self, med, name, cfg=None, variable_names=None, **kwargs):
        """Convenience constructor for default geographic data source."""
        variable_names = variable_names or [GEO_VARIABLE_NAME]
        super(GeoDataSourceBase, self).__init__(
            med, name, cfg=cfg, variable_names=variable_names,
            data_as_dict=True, **kwargs)

        #: Area and zones/regions description per data sources.
        self.area_places_sources = None

        #: Area-zones/regions mapping for data source.
        self.area_places = None

        #: Area-regions mapping for data source.
        self.area_regions = None

        #: Place-area mapping.
        self.place_area = None

        #: Areas.
        self.areas = None

        #: Region-zone/region mapping.
        self.region_place = None

        #: Zone/region-region mapping.
        self.place_regions = None

        #: Data-source region names.
        self.region_names = None

        #: Zone/region names.
        self.place_names = None

        #: Zone/region indices = None
        self.place_indices = None

        # Get mappings and list
        self._init_regions_places(**kwargs)

        #: Child column.
        if hasattr(self, 'cfg') and getattr(self, 'cfg'):
            self._child_column = self.cfg.get('child_column')

    def get_place_regions_for_source(self, src_name, area=None, **kwargs):
        """Get zone/region-regions and region-zone/region mappings
        for data source and area.

        :param src_name: Data-source name for which to get place names.
        :param area: Area for which to get place names. Default is `None`,
          in which case places for all areas are returned.
        :type src_name: str
        :type area: str

        :returns: zone/region-regions and region-zone/region mappings.
        :rtype:

          * :py:class:`dict` of :py:class:`str` to :py:class:`set`
          * :py:class:`dict` of :py:class:`str` to :py:class:`str`

        """
        # Get place names for given area but all data sources
        area_place_names = (self.area_places[area] if area is not None else
                            self.place_names)
        area_place_names = ensure_collection(area_place_names, set)
        areas = [area] if area is not None else self.areas

        # Get place-regions mapping for given area and data source
        src_area_place_regions = OrderedDict()
        for area in areas:
            src_area_place_regions.update(
                {place_name: ensure_collection(src_regions[src_name])
                 for place_name, src_regions
                 in self.area_places_sources[area].items()
                 if src_name in src_regions})

        # Get region-place mapping
        src_area_region_place = OrderedDict()
        for place_name, region_names in src_area_place_regions.items():
            src_area_region_place.update({
                region_name: place_name for region_name in region_names})

        return src_area_place_regions, src_area_region_place

    def get_total_bounds(self, epsg=4326, **kwargs):
        """Get array of min and max coordinates on each axis.

        :param epsg: EPSG code specifying output projection.
        :type epsg: int

        :returns: Array with min x, min y, max x, max y.
        :rtype: :py:class:`numpy.array`
        """
        # Get geo data
        geo_data = self.med.geo_src.get_data(
            variable_names=GEO_VARIABLE_NAME, **kwargs)[GEO_VARIABLE_NAME]

        # Transform to right CRS
        data = geo_data.to_crs(epsg=epsg) if epsg else geo_data

        # Return total bounds
        return data.total_bounds

    def read_place_coordinates(self, **kwargs):
        """Read centroid coordinates of zones/regions.

        :returns: Centroid coordinates of places.
        :rtype: :py:class:`pandas.GeoDataFrame`
        """
        # Set filepath
        file_dir = self.med.cfg.get_project_data_directory(self, **kwargs)
        filename = 'latlon{}.csv'.format(
            self.get_data_postfix(with_src_name=True, **kwargs))
        filepath = os.path.join(file_dir, filename)

        # Open and read coordinates from file
        with open(filepath, 'r') as f:
            df_coord_places = pd.read_csv(f, index_col=0)

        return df_coord_places

    def write_place_coordinates(self, **kwargs):
        """Write centroid coordinates of zones/regions."""
        # Get coordinates as data frame
        gdf = self.get(GEO_VARIABLE_NAME)
        df_coord = pd.DataFrame(index=gdf.index, columns=['lat', 'lon'])

        for place in gdf.index:
            # Get place centroid in source coordinates
            pt0 = gdf.loc[place].geometry.centroid

            # Convert to reference CRS
            pt = gpd.GeoSeries(pt0, crs=gdf.crs).to_crs(
                from_epsg(4326)).values[0]
            df_coord.loc[place] = [pt.y, pt.x]

        # Set filepath
        file_dir = self.med.cfg.get_project_data_directory(self, **kwargs)
        filename = 'latlon{}.csv'.format(
            self.get_data_postfix(with_src_name=True, **kwargs))
        filepath = os.path.join(file_dir, filename)

        # Write coordinates to file
        df_coord.to_csv(filepath)

    def _init_regions_places(self, **kwargs):
        """Set region-to-zone/region mapping :py:attr:`region_place`,
        zone/region-names list :py:attr:`place_names` and
        zone/region-to-regions mapping :py:attr:`place_regions`.

        .. note:: :py:attr:`areas` are to areas in
          `component_managers_per_area` of mediator configuration, if defined,
          or to all areas in :py:attr:`area_places_sources`.

        .. seealso:: :py:meth:`_get_area_places_sources`.
        """
        # Get area and zones/regions description per data sources
        self.area_places_sources = self._get_area_places_sources(**kwargs)

        # Get all areas from components or all described areas
        areas = set(self.med.cfg.get('component_managers_per_area') or
                    self.area_places_sources)

        # Loop over areas to define places and regions
        self.area_places = OrderedDict()
        self.area_regions = OrderedDict()
        self.region_place = OrderedDict()
        self.place_regions = OrderedDict()
        self.place_names = set()
        self.areas = set()
        for area in areas:
            # Loop over place-source names pairs
            place_sources = self.area_places_sources[area]
            for place_name, src_regions in place_sources.items():
                # Try to get region names for this data source
                if self.name in src_regions:
                    # Add area
                    self.areas.add(area)

                    # Add place to place names
                    self.place_names.add(place_name)

                    # Add place to area
                    self.area_places.setdefault(area, set()).add(place_name)

                    # Get place region-names.
                    region_names = ensure_collection(
                        src_regions[self.name], set)

                    # Add regions to area
                    self.area_regions.setdefault(area, set()).update(
                        region_names)

                    # Add place-regions pair
                    self.place_regions[place_name] = ensure_collection(
                        region_names)

                    # Add region-place pair
                    self.region_place.update(
                        {region_name: place_name
                         for region_name in region_names})

        # Get place-area mapping
        self.place_area = OrderedDict()
        for area, place_names in self.area_places.items():
            self.place_area.update(
                {place_name: area for place_name in place_names})

        # Get data-source region names
        self.region_names = set(self.region_place)

        # Get place indices
        self.place_indices = self._get_place_indices(**kwargs)

    def _get_area_places_sources(self, **kwargs):
        """Get zones/regions from file given by `places_filepath` entry of
        geo configuration.

        :returns: Places.
        :rtype: list
        """
        area_places_sources_filepath = self.med.geo_cfg.get('places_filepath')
        area_places_sources_filepath = os.path.join(*ensure_collection(
            area_places_sources_filepath))
        with open(area_places_sources_filepath, 'r') as f:
            area_places_sources = yaml.load(f, Loader=yaml.FullLoader)

        return area_places_sources

    def _get_place_indices(self, **kwargs):
        """Get place indices in mask."""
        # Create dictionary of places (independent of data grid)
        place_ids = list(range(2, len(self.place_names) + 2))
        place_indices = dict(zip(self.place_names, place_ids))

        return place_indices


class MultiGeoDataSource(MultiDataSource, GeoDataSourceBase):
    """Multiple geographic data source.
    Redefines :py:meth:`get` method to merge (geographic) variable
    in all single data sources into a single
    :py:class:`pandas.GeoDataFrame`."""

    def __init__(self, med, data_sources, task_names=set(),
                 default_tasks_value=True, **kwargs):
        """Build with :py:class:`.data_source.MultiDataSource` constructor.

        :param med: Mediator.
        :param data_sources: Data sources dictionary.
        :param task_names: Names of potential tasks for container to perform.
          Default is `set()`.
        :param default_tasks_value: If `True`, ask to perform all tasks.
          Otherwise, none. Default is `True`.
        :type med: :py:class:`.mediator.Mediator`
        :type data_sources: :py:class:`dict` of :py:class:`DataSourceBase`
        :type task_names: set
        :type default_tasks_value: bool
        """
        # Initialize as data source
        super(MultiGeoDataSource, self).__init__(
            med, data_sources, task_names=set(),
            default_tasks_value=True, **kwargs)

        # Initialize regions and places merging from single data sources
        self._init_regions_places(**kwargs)

    def _init_regions_places(self, **kwargs):
        """Set region-to-zone/region mapping :py:attr:`region_place`,
        zone/region-names set :py:attr:`place_names` and
        zone/region-to-regions mapping :py:attr:`place_regions`
        merging from single data sources.
        """
        # Make sure that the initialization is called after
        # the parents constructors are called
        if hasattr(self, 'data_sources') and getattr(self, 'data_sources'):
            # Get zones/regions description from any single data source
            self.area_places_sources = self.data_sources[
                list(self.data_sources)[0]].area_places_sources

            # Merge attributes
            self.region_place = OrderedDict()
            self.place_regions = OrderedDict()
            self.place_indices = OrderedDict()
            self.area_places = OrderedDict()
            self.place_area = OrderedDict()
            self.place_names = set()
            self.region_names = set()
            self.areas = set()
            for data_src in self.data_sources.values():
                # Merge mappings
                self.region_place.update(data_src.region_place)
                self.place_regions.update(data_src.place_regions)
                self.place_indices.update(data_src.place_indices)
                self.place_area.update(data_src.place_area)
                # Allow for multiple data sources per area
                for area, places in data_src.area_places.items():
                    self.area_places.setdefault(area, set()).update(places)

                # Merge sets
                self.place_names.update(data_src.place_names)
                self.region_names.update(data_src.region_names)
                self.areas.update(data_src.areas)

    def get(self, variable_name, default=None):
        """Get variable from data source containing variable.

        :param variable_name: Variable name.
        :param default: Default geographic data frame. Default is `None`.
        :type variable_name: str
        :type default: :py:class:`pandas.GeoDataFrame`

        :returns: Merged geographic data frame.
        :rtype: py:class:`pandas.GeoDataFrame`
        """
        return self.data.get(variable_name, default)

    def get_data(self, **kwargs):
        """Load geographic data from multiple geographic data sources
        and merge the geographic data into one geographic data frame.


        :returns: Dataset :py:attr:`data`.
        :rtype: mapping of :py:class:`str` to :py:class:`pandas.GeoDataFrame`
        """
        gdf = None

        # Get reference CRS
        data_sources = self.data_sources.copy()
        crs_as = self.med.geo_cfg.get('crs_as')
        if crs_as:
            data_src = data_sources.pop(crs_as)
            gdf = data_src.get_data(**kwargs)[GEO_VARIABLE_NAME]
            crs = gdf.crs

        for data_src in data_sources.values():
            # Get data for geo variable of single source
            gdf_single = data_src.get_data(**kwargs)[GEO_VARIABLE_NAME]

            if crs_as:
                # Convert to reference CRS
                gdf_single = gdf_single.to_crs(crs)

            # Merge geographic data frame
            gdf = gdf_single if gdf is None else pd.concat([gdf, gdf_single])

        # Assign
        self.data = {GEO_VARIABLE_NAME: gdf}

        return self.data


class DefaultMaskAPI(DataSourceLoaderBase):
    """Mask downloading mixin. To be mixed with
    :py:class:`GeoDataSourceBase`.
    """
    #: Flag allowing several downloads per area.
    ONE_DOWNLOAD_PER_AREA = False

    def download(self, **kwargs):
        """Download geographic data defining geometries of areas/zones/regions.

        :returns: Names of downloaded variables.
        :rtype: :py:class:`set` of :py:class:`str`

        .. warning:: Requires :py:meth:`get_url` and :py:meth:`get_filename`
          methods to be implemented.
        """
        src_dir = self.med.cfg.get_external_data_directory(self, **kwargs)
        previous_urls = []
        for area in self.areas:
            # Get URL and filename for area
            url = self.get_url(area=area, **kwargs)
            filename = self.get_filename(area=area, **kwargs)

            # Prevent downloading non-area specific data multiple times
            if url not in previous_urls:
                # Download geographic data for area
                self.log.info(
                    'Downloading {} geographic data for {} from {}'.format(
                        self.name, area, url))
                download_from_url(url, filename, src_dir)

                if self.ONE_DOWNLOAD_PER_AREA:
                    # Prevent future downloads
                    previous_urls.append(url)

        return {GEO_VARIABLE_NAME}

    def read_file(self, area, **kwargs):
        """Read downloaded geographical data for area.

        :param area: Geographical area for which to read the data.
        :type area: str

        :returns: Geographic data frame.
        :rtype: :py:class:`geopandas.GeoDataFrame`
        """
        src_dir = self.med.cfg.get_external_data_directory(
            self, makedirs=False, **kwargs)

        # Get directory for area
        filename = self.get_filename(area=area, **kwargs)
        filepath = os.path.join(src_dir, filename)

        # Read geographical file for area
        read_file_kwargs = self.cfg.get('read_file_kwargs') or {}
        gdf_area = gpd.read_file(filepath, **read_file_kwargs)

        return gdf_area

    def load(self, **kwargs):
        """Load geographical data for all areas.

        :returns: Geographical array with geometries.
        :rtype: mapping to :py:class:`geopandas.GeoDataFrame`

        .. note:: The geographic data is aggregated by zone.
          A zone could be the region itself,
          in which case no aggregation is performed.

        """
        gdf = None
        for area in self.areas:
            # Read file
            gdf_area = self.read_file(area, **kwargs)

            # Handle parent/child columns
            if 'parent_column' in self.cfg:
                # Get country code
                country_code = get_country_code(area, code='alpha-2')

                # Select country
                gdf_area_sel = gdf_area[gdf_area.loc[
                    :, self.cfg['parent_column']] == country_code]

                # GB/UK exception
                if (country_code == 'GB') and (len(gdf_area_sel) == 0):
                    gdf_area_sel = gdf_area[gdf_area.loc[
                        :, self.cfg['parent_column']] == 'UK']

                gdf_area = gdf_area_sel

            # Index by child-column
            gdf_area = gdf_area.set_index(self._child_column)[['geometry']]

            region_names = self.area_regions[area]
            if (len(region_names) == 1) and (area in region_names):
                # Manage country case
                gdf_area.loc[:, 'zone'] = area
                gdf_area = gdf_area.dissolve(by='zone')
            else:
                # Select requested regions for area
                gdf_area = gdf_area.loc[region_names]

                # Aggregate by zone taking places from region-place mapping
                # to preserve order.
                gdf_area.loc[:, 'zone'] = [self.region_place[region_name]
                                           for region_name in gdf_area.index]
                gdf_area = gdf_area.dissolve(by='zone')

            # Merge area
            gdf = gdf_area if gdf is None else pd.concat([gdf, gdf_area])

        # Add variable data to data source
        self.update({GEO_VARIABLE_NAME: gdf})

        # Write coordinates of places' centroids
        self.write_place_coordinates(**kwargs)

        return {GEO_VARIABLE_NAME: gdf}

    def read(self, **kwargs):
        """Read source dataset as :py:class:`geopandas.GeoDataFrame`
        from file.
        """
        filepath = self.get_data_path(
            variable_name=GEO_VARIABLE_NAME, makedirs=False, is_timeseries=False,
            **kwargs)
        if not self.cfg.get('no_verbose'):
            self.log.info('Reading {} for {} from {}'.format(
                GEO_VARIABLE_NAME, self.name, filepath))
        self[GEO_VARIABLE_NAME] = gpd.read_file(filepath).set_index('zone')

    def write(self, **kwargs):
        """Write source :py:class:`geopandas.GeoDataFrame` dataset to file.
        """
        filepath = self.get_data_path(
            variable_name=GEO_VARIABLE_NAME, is_timeseries=False, **kwargs)
        if not self.cfg.get('no_verbose'):
            self.log.info('Writing {} for {} to {}'.format(
                GEO_VARIABLE_NAME, self.name, filepath))
        self.get(GEO_VARIABLE_NAME).reset_index().to_file(filepath)


class MaskMakerBase(ABC):
    """Base mask maker. To be mixed with :py:class:`GeoDataSourceBase`.
    ..warning:: Requires :py:meth:`make_mask` method to be implemented.
    """
    @abstractmethod
    def make_mask(self, *args, **kwargs):
        """Make mask for a gridded data source."""
        raise NotImplementedError


class DefaultMaskMaker(GeoDataSourceBase, MaskMakerBase):
    """Default mask maker."""

    def make_mask(self, data_src, **kwargs):
        """Make mask for a given gridded data source, store the
        regions' geometries to :py:attr:`data` member.

        :param data_src: Gridded data source.
        :type data_src: :py:class:`.grid.GriddedDataSourceBase`

        :returns: Mask dataset.
        :rtype: :py:class:`xarrray.Dataset`
        """
        # Get regions' geometries
        self.get_data(variable_names=GEO_VARIABLE_NAME, **kwargs)

        # Download data to read the grid
        data_src.manage_download(**kwargs)

        # Get data coordinates
        coords = data_src.get_grid(**kwargs)

        # Get point region membership
        self.log.info('Assigning points to regions')
        ds_mask = get_point_region_membership(
            self.cfg, self.get(GEO_VARIABLE_NAME), self.place_indices, coords)

        # Add geographic data-source name
        ds_mask.attrs['data_source'] = self.name

        return ds_mask


def get_point_region_membership(cfg_src, gdf, place_indices, coords):
    """Assign grid-points of dataset to electricity places.

    :param cfg_src: Data source configuration.
    :param gdf: Regions' geometries.
    :param place_indices: Dictionary assigning each plae name to a place ID.
      IDs of places of interest should be larger or equal to 2.
    :param coords: Input data coordinates assigned to telectricity places.
    :type cfg_src: dict
    :type gdf: :py:class:`geopandas.GeoDataFrame`
    :type place_indices: dict
    :type coords: :py:class:`tuple` of pairs of :py:class:`str`
      and :py:class:`numpy.array`

    :returns: Dataset containing mask assigning each grid-point to a place.
    :rtype: :py:class:`xarray.Dataset`
    """
    # Get dims and coords for regular and irregular grids
    try:
        dim_names = coords.dims
        dims = tuple(coords[dim].shape[0] for dim in dim_names)
    except AttributeError:
        dim_names = coords.keys()
        dims = tuple(len(v) for v in coords.values())

    # Create an empty mask
    ds = xr.Dataset()
    ds['mask'] = xr.DataArray(np.zeros(dims, dtype=int),
                              dims=dim_names, coords=coords)

    # Add region indices
    coords = ('region', list(place_indices.keys()))
    ds['region_index'] = xr.DataArray(
        list(place_indices.values()), coords=[coords])

    # Assign points to regions
    for iy in range(ds['mask'].shape[0]):
        for ix in range(ds['mask'].shape[1]):
            # Select point and convert to regions' CRS
            try:
                lat = float(ds.lat[iy, ix].values)
                lon = float(ds.lon[iy, ix].values)
            except IndexError:
                lat = float(ds.lat[iy].values)
                lon = float(ds.lon[ix].values)
            pt = gpd.GeoSeries(
                Point(lon, lat), crs=from_epsg(4326)).to_crs(gdf.crs)

            # Select place to which this point belongs to, if any
            within = gdf.contains(pt[0])
            if within.any():
                # Get place name
                zone = gdf[within].index[0]

                # Get corresponding zone and save zone ID in mask
                # If region in no zone, set to 1
                ds['mask'][iy, ix] = place_indices[zone] if zone else 1

    # Remove empty regions
    idx_filled = np.in1d(ds.region_index, np.unique(ds.mask))
    ds = ds.loc[{'region': idx_filled}]

    return ds


def download_from_url(url, filename, src_dir):
    """Download geographic data defining borders for areas/zones/regions
    and return geometries.

    :param url: URL from which to download data.
    :param filename: Downloaded-file name.
    :param src_dir: Data destination directory.
    :type url: str
    :type filename: str
    :type src_dir: str

    .. note:: This function is not directly called from this module,
      but from API's included in this package(e.g. GISCO).
    """
    # Download data for region
    filepath = os.path.join(src_dir, filename)
    r = requests.get(url)
    if r.status_code != 200:
        raise FileNotFoundError(url)

    # Extract, if needed, or write file
    if url[-4:] == '.zip':
        zip_ref = zipfile.ZipFile(BytesIO(r.content))
        zip_ref.extractall(src_dir)
        zip_ref.close()
    else:
        with open(filepath, 'wb') as f:
            for chunk in r:
                f.write(chunk)


def get_country_code(area, code='alpha-2'):
    """Get country code.

    :param area: Country name.
    :param code: Code name.
    :type area: str
    :type code: str

    :returns: Country code.
    :rtype: str
    """
    # Read country codes
    resource_name = 'data/iso_country_codes.csv'
    with resource_stream(__name__, resource_name) as stream:
        cc_data = pd.read_csv(stream, index_col=0)
    area_name = difflib.get_close_matches(
        area, cc_data.index.tolist(), n=1)[0]

    return cc_data.loc[area_name, code]



"""Electricity-network power-flow problems.

  .. todo:: Adapt power-flow problems to new version.
"""
import numpy as np


def get_transmission_dc(sus_mat, delta):
    """ Compute the transmission matrix from the susceptance
    matrix and the phases computed beforehand from the DC approximation
    of the power flow.

    :param sus_mat: Susceptance matrix,
      i.e. the imaginary part of the admittance matrix.
    :param delta: Phase at each node.
    :type sus_mat: :py:obj:`numpy.array`
    :type delta: :py:obj:`numpy.array`

    :returns: Transmission matrix.
    :rtype: :py:obj:`numpy.array`
    """
    n_reg = sus_mat.shape[0]
    trans_mat = np.zeros((n_reg, n_reg))
    for i in np.arange(n_reg):
        for j in np.arange(i + 1, n_reg):
            trans_mat[i, j] = (delta[i] - delta[j]) * sus_mat[i, j]
            trans_mat[j, i] = -trans_mat[i, j]

    return trans_mat


def power_flow_dc(power_trans, sus_mat):
    """ Solve the power flow problem using the DC approximation.

    :param power_trans: Power to be transmitted at each node of the network.
    :param sus_mat: The susceptance matrix (S),
      i.e. the imaginary part of the admittance matrix).
    :type power_trans: :py:obj:`numpy.array`
    :type sus_mat: :py:obj:`numpy.array`

    :returns: A tuple containing:

      * The transmission between each pair of node.
      * The phase at each node.

    :rtype: tuple

    .. note::
      Node one is taken as slack node. All the other nodes are
      considered generation nodes.

    .. see also::
      * Grainger, J., Stevenson, Jr. W., 1994, *Power System Analysis*,
        McGraw-Hill, New York.
      * Glover, J. D., Sarma, M. S., Overbye, T., 2012,
        *Power System Analysis and Design*, CENGAGE Learning.
    """
    nt, n_reg = power_trans.shape

    # Solve the power flow for each time step
    delta = np.zeros((nt, n_reg))
    trans = np.empty((nt, n_reg, n_reg))
    for t in np.arange(nt):
        # Solve the power flow to find the phases
        delta[t, 1:] = np.linalg.solve(-sus_mat[1:, 1:], power_trans[t, 1:])

        # Compute the resulting transmission between each pair of node.
        trans[t] = get_transmission_dc(sus_mat, delta[t])

    return (trans, delta)



"""CBS API for the Netherlands."""
import os
import requests
import pandas as pd
import xarray as xr
from ..data_source import DataSourceLoaderBase
from ..component import (
    DataSourceWithComponentsMixin, parse_variable_component_args,
    download_to_compute_capacity_factor, compute_capacity_factor,
    finalize_dataset)


class DataSource(DataSourceLoaderBase, DataSourceWithComponentsMixin):
    """cbs.nl data source."""
    #: Default source name.
    DEFAULT_SRC_NAME = 'cbs'

    #: Area: `'Netherlands'`
    AREA = 'Netherlands'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    @parse_variable_component_args
    @download_to_compute_capacity_factor
    def download(self, variable_component_names, **kwargs):
        """Download data.

        :param variable_component_names: Names of components to download per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to :py:class:`set` of :py:class:`str`

        :returns: Names of downloaded components per variable.
        :rtype: mapping from :py:class:`str` to :py:class:`set` of
          :py:class:`str`

        .. note:: The data is downloaded from
          `CBS <https://opendata.cbs.nl/statline/portal.html>`.
        """
        # Loop over variables
        for variable_name, component_names in variable_component_names.items():
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Loop over components
            for component_name in component_names:
                # Get URL
                url = self._get_url(
                    variable_name, component_name, **kwargs)
                filepath = self._get_download_filepath(
                    variable_name, component_name, **kwargs)
                if not self.cfg.get('no_verbose'):
                    self.log.info('--- {} from {} to {}'.format(
                        component_name, url, filepath))

                # Request and raise exception if needed
                response = requests.get(url)
                response.raise_for_status()

                # Make data frame
                records = response.json()['value']
                df = pd.DataFrame(records).set_index('Perioden')
                index = ('{}-12-31'.format(t[:4]) for t in df.index)
                df.index = pd.DatetimeIndex(index)
                df.index.name = 'time'

                # Write data frame
                df.to_csv(filepath)

        # Return names of downloaded variables
        return variable_component_names

    @parse_variable_component_args
    @compute_capacity_factor
    @finalize_dataset
    def load(self, variable_component_names, **kwargs):
        """Load data.

        :param variable_component_names: Names of components to load per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to collection

        :returns: Time series for each variable and component.
        :rtype: :py:class:`xarray.Dataset()`
        """
        # Get place names for source
        src_place_regions, _ = (
            self.med.geo_src.get_place_regions_for_source(
                self.name, self.AREA, **kwargs))
        place_names = list(src_place_regions)

        # Loop over variables
        ds = xr.Dataset()
        for variable_name, component_names in variable_component_names.items():
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Loop over components
            da = None
            for component_name in component_names:
                # Get filepath
                filepath = self._get_download_filepath(
                    variable_name, component_name, **kwargs)
                if not self.cfg.get('no_verbose'):
                    self.log.info('-- {} from {}'.format(
                        component_name, filepath))

                # Read data frame
                df_comp = pd.read_csv(
                    filepath, **self.cfg['read_csv_kwargs'])
                df_comp.columns = [component_name]

                # Convert to data array
                da_comp = xr.DataArray(df_comp, dims=('time', 'component'))

                # Add component to data array
                da = (da_comp if da is None else
                      xr.concat([da, da_comp], dim='component'))

            # Add area as region coordinate
            da = da.expand_dims('region').assign_coords(
                region=place_names)

            # Add variable to dataset
            ds[variable_name] = da

        return ds

    def _get_capacity_factor(self, ds, **kwargs):
        """Get capacity factor from generation and capacity.

        :param ds: Dataset containing generation and capacity.
        :type ds: mapping

        :returns: Capacity factor.
        :rtype: :py:class:`xarray.DataArray`
        """
        da_gen = ds['generation']

        # Get number of days in month
        time = da_gen.indexes['time']
        start = pd.Timestamp(time[0])
        end = pd.Timestamp(time[-1]) + pd.Timedelta('366 days')
        time_plus = pd.date_range(start=start, end=end, freq='Y')
        ndays = (time_plus[1:] - time_plus[:-1]).days
        da_ndays = xr.DataArray(ndays, coords=[('time', time)])
        da_nhours = da_ndays * 24

        # Get capacity factor
        da = da_gen / (ds['capacity'] * da_nhours)

        return da

    def _get_url(self, variable_name, component_name, **kwargs):
        """Get URL for query.

        :param variable_name: Name of variable to query.
        :param component_name: Name of component to query.
        :type variable_name: str
        :type component_name: str

        :returns: URL.
        :rtype: str
        """
        # Filter component
        filter_component = "BronTechniek eq '{}'".format(
            self.cfg['component_keys'][component_name])

        # Filter periods
        filter_time = ("substring(Perioden, 0, 4) ge '{}' and "
                       "substring(Perioden, 0, 4) le '{}'"
                       "substring(Key,4,2) eq 'JJ'".format(
                           self.cfg['first_year'], self.cfg['last_year']))

        # Compile filters
        filters = '{} and {}'.format(filter_component, filter_time)

        # Select properties
        select = 'Perioden,{}'.format(
            self.cfg['variable_names'][variable_name])

        # Define URL
        url = '{}/{}/{}?$filter={}&$select={}'.format(
            self.cfg['host'], self.cfg['variable_identifiers'][variable_name],
            self.cfg['data_entity'], filters, select)

        return url

    def _get_download_filepath(self, variable_name, component_name, **kwargs):
        """Get download filepath.

        :param variable_name: Name of variable to query.
        :param component_name: Name of component to query.
        :type variable_name: str
        :type component_name: str

        :returns: Filepath.
        :rtype: str
        """
        # Define filename
        filename = '{}_{}_{}{}.csv'.format(
            self.name, variable_name, component_name, self.get_postfix())

        # Get filepath
        src_dir = self.med.cfg.get_external_data_directory(self)
        filepath = os.path.join(src_dir, filename)

        return filepath

    def get_postfix(self, **kwargs):
        """Get data-source standard postfix.

        :returns: Postfix.
        :rtype: str
        """
        postfix = '_{}-{}'.format(self.cfg['first_year'],
                                  self.cfg['last_year'])

        return postfix



"""Various plot functions."""
import os
import numpy as np
import cartopy.crs as ccrs
import matplotlib.pyplot as plt
from matplotlib import rcParams
from pandas.plotting import register_matplotlib_converters
from .geo import GEO_VARIABLE_NAME
from .container import load_config

# Register matplotlib datetime-converter
register_matplotlib_converters()

#: Default plot-colors.
COLORS = rcParams['axes.prop_cycle'].by_key()['color']


def plot_mask(data_src, crs=None, facecolor='moccasin', edgecolor='dimgrey',
              **kwargs):
    """ Plot the mask assigning climate grid points to regions.

    :param data_src: Data source.
    :param crs: Coordinate Reference System. Default is
      :py:class:`cartopy.crs.LambertAzimuthalEqualArea`.
    :param facecolor: Face color of the regions. Default is `'moccasin'`.
    :param edgecolor: Edge color of the regions. Default is `'dimgrey'`.
    :type data_src: :py:class:`.data_source.DataSourceBase`
    :type crs: :py:class:`cartopy.crs.ABCMeta`
    :type facecolor: str
    :type edgecolor: str
    """
    # Configuration
    med = data_src.med
    cfg_plot = load_config(med, 'plot')
    fig_format = cfg_plot['savefigs_kwargs']['format']
    fig_dir = med.cfg.get_plot_directory(data_src, **kwargs)
    markersize = 5
    facecolor = facecolor or cfg_plot.get('region_facecolor') or 'moccasin'
    edgecolor = edgecolor or cfg_plot.get('region_edgecolor') or 'dimgrey'
    crs = crs or ccrs.LambertAzimuthalEqualArea()

    # Get mask
    data_src.get_mask(**kwargs)
    mask = data_src.mask

    # Convert geometry to CRS
    gdf = med.geo_src.get_data(variable_names=GEO_VARIABLE_NAME)[
        GEO_VARIABLE_NAME]
    gdf_crs = gdf.to_crs(crs.proj4_init)

    # Plot
    if not cfg_plot.get('no_verbose'):
        med.log.info('Plotting {} mask for {}'.format(
            med.geo_src.name, data_src.name))
    fig = plt.figure()
    ax = plt.axes(projection=crs)
    ax.set_xlabel('Longitude', fontsize=cfg_plot['fs_default'])
    ax.set_ylabel('Latitude', fontsize=cfg_plot['fs_default'])

    # Plot regions
    gdf_crs.plot(ax=ax, facecolor=facecolor, edgecolor=edgecolor)

    # Get grid points within regions
    ilat_in, ilon_in = np.where(mask['mask'] > 1)

    for ilat, ilon in zip(ilat_in, ilon_in):
        place_id = int(mask['mask'][ilat, ilon])
        if len(mask.lat.dims) > 1:
            lat, lon = mask.lat[ilat, ilon], mask.lon[ilat, ilon]
        else:
            lat, lon = mask.lat[ilat], mask.lon[ilon]
        ax.scatter(lon, lat, s=markersize,
                   transform=ccrs.Geodetic(),
                   c=COLORS[place_id % len(COLORS)])
    mask_in = mask.mask[ilat_in, ilon_in]
    ax.set_extent([mask_in.lon.min(), mask_in.lon.max(),
                   mask_in.lat.min(), mask_in.lat.max()])
    ax.set_title('{} {}'.format(med.geo_src.name, data_src.name))

    fig_filename = 'mask{}.{}'.format(data_src.get_mask_postfix(), fig_format)
    fig_filepath = os.path.join(fig_dir, fig_filename)
    if not cfg_plot.get('no_verbose'):
        med.log.info('Saving figure for {} mask for {} to {}'.format(
            med.geo_src.name, data_src.name, fig_filepath))
    fig.savefig(fig_filepath, **cfg_plot['savefigs_kwargs'])

    if cfg_plot['show']:
        plt.show(block=False)


def plot_data_source(data_src, per_region=None, **kwargs):
    """ Plot data from a given source.

    :param data_src: Data source.
    :param per_region: If `True`, plot on a separate figure for each region.
      If `None`, try reading it from mediator configuration.
      Default is `False`.
    :type data_src: :py:class:`.data_source.DataSourceBase`
    :type per_region: bool

    .. seealso:: :py:func:`plot_regional_dataset`
    """
    fig_dir = data_src.med.cfg.get_plot_directory(data_src, **kwargs)
    info_msg = ' {}'.format(data_src.name)

    # Get data
    data_src.get_data(**kwargs)

    # Plot dataset
    for variable_name in data_src.variable_names:
        plot_regional_dataset(
            data_src, variable_name, fig_dir=fig_dir, info_msg=info_msg,
            per_region=per_region, **kwargs)


def plot_result(result_mng, result_name='result',
                stage=None, per_region=None, **kwargs):
    """ Plot features of a given result manager.

    :param result_mng: Result manager of variable for which to plot
      features.
    :param result_name: `'feature'` or `'prediction'`.
    :param stage: Modeling stage: `'fit'` or `'predict'`.
      Should be provided for `'fit'`. Default is `None`.
    :param per_region: If `True`, plot on a separate figure for each region.
      If `None`, try reading it from mediator configuration.
      Default is `False`.
    :type result_mng: :py:class:`.component.ResultManager`
    :type result_name: str
    :type stage: str

    .. seealso:: :py:func:`plot_regional_dataset`
    """
    # Get data
    info_msg = ' {}'.format(result_name)
    if result_name == 'feature':
        # Plot feature
        # Make sure that result is loaded and get data
        result_mng.extract(stage, **kwargs)
        data_src = result_mng.feature[stage]
        if not data_src.data:
            return
        info_msg += ' to {}'.format(stage)
    elif result_name == 'result':
        # Plot prediction or (extracted) output
        # Make sure that result is loaded and get data
        result_mng.get_data(**kwargs)

    info_msg += ' {}'.format(result_mng.component_mng.name)
    fig_dir = result_mng.med.cfg.get_plot_directory(
        result_mng.component_mng, **kwargs)

    for variable_name in result_mng.variable_names:
        plot_regional_dataset(
            result_mng, variable_name, fig_dir=fig_dir, info_msg=info_msg,
            per_region=per_region, **kwargs)


def plot_regional_dataset(data_src, variable_name, fig_dir='', info_msg='',
                          per_region=None, **kwargs):
    """Plot all variable of a regional dataset.

    :param data_src: Dataset.
    :param variable_name: Variable name.
    :param fig_dir: Directory in which to save figure.
    :param info_msg: Log and title information message.
    :param per_region: If `True`, plot on a separate figure for each region.
      If `None`, try reading it from mediator configuration.
      Default is `False`.
    :type data_src: mapping
    :type variable_name: str
    :type fig_dir: str
    :type info_msg: str
    :type per_region: bool
    """
    med = data_src.med
    cfg_plot = load_config(med, 'plot')
    fig_format = cfg_plot['savefigs_kwargs']['format']
    per_region = per_region or cfg_plot.get('per_region', False)

    if not cfg_plot.get('no_verbose'):
        med.log.info('Plotting {} {}'.format(info_msg, variable_name))
    var = data_src.get_data()[variable_name]
    if not cfg_plot.get('no_verbose'):
        med.log.info('Variable: {}'.format(variable_name))

    # Define y label with units
    units = var.attrs.get('units')
    sunits = ' ({})'.format(units) if (units and (units != '')) else ''
    ylabel = '{}{}'.format(variable_name, sunits)

    # Plot per region
    if not hasattr(var, 'region'):
        # Check if regional data
        var = var.expand_dims(dim='region', axis=-1).assign_coords(
            region=['all'])
    time = var.indexes['time']
    freq_data = time.inferred_freq
    if freq_data is None:
        if (time[1] - time[0]).seconds == 3600:
            freq_data = 'H'
    freq_data = freq_data.upper()
    if ((med.cfg['frequency'] == 'day') and
            (freq_data in ['H', '1H'])):
        var = var.resample(time='D').mean('time')
    for ir, reg_label in enumerate(var.indexes['region']):
        plot_postfix = '_{}'.format(reg_label) if per_region else ''
        reg_msg = ' - {}'.format(reg_label) if per_region else ''
        if per_region or (ir == 0):
            fig, ax = plt.subplots(1, 1)
            da = var.loc[{'region': reg_label}]
            time = da.indexes['time']
            try:
                for component_name in var.indexes['component']:
                    label = '{} {}'.format(component_name, reg_label)
                    ax.plot(time, da.loc[{'component': component_name}],
                            label=label)
            except KeyError:
                ax.plot(time, da, label=reg_label)

        last = (ir == len(var.indexes['region']) - 1)
        if not per_region and last:
            ax.legend(loc='best')
        if per_region or last:
            # Set axes
            ax.set_xlim(time[0], time[-1])
            ax.set_xlabel('time', fontsize=cfg_plot['fs_default'])
            ax.set_ylabel(ylabel, fontsize=cfg_plot['fs_default'])
            ax.set_title('{}{} {}'.format(reg_msg, info_msg, variable_name))

            # Save figure
            result_postfix = data_src.get_data_postfix(
                variable_name=variable_name, **kwargs)
            fig_filename = '{}{}{}.{}'.format(
                variable_name, result_postfix, plot_postfix, fig_format)
            fig_filepath = os.path.join(fig_dir, fig_filename)
            if not cfg_plot.get('no_verbose'):
                med.log.info('Saving figure to {}'.format(fig_filepath))
            fig.savefig(fig_filepath, **cfg_plot['savefigs_kwargs'])

    if cfg_plot['show']:
        plt.show(block=False)


def plot_generation_feature(result_mng, stage=None, **kwargs):
    """Plot generation features of a given result manager.

    :param result_mng: Result manager for which to plot features.
    :param stage: Modeling stage: `'fit'` or `'predict'`.
      Should be provided Default is `None`.
    :type result_mng: :py:class:`.component.ResultManager`
    :type stage: str
    """
    # Plot
    med = result_mng.med
    cfg_plot = load_config(med, 'plot')
    fig_format = cfg_plot['savefigs_kwargs']['format']
    fig_dir = med.cfg.get_plot_directory(
        result_mng.component_mng, **kwargs)
    os.makedirs(fig_dir, exist_ok=True)

    if not cfg_plot.get('no_verbose'):
        med.log.info('Plotting features to {} {}'.format(
            stage, result_mng.name))

    # Get data
    result_mng.extract(stage)
    data_src = result_mng.feature[stage]
    if not data_src.data:
        return
    result_postfix = data_src.get_data_postfix(**kwargs)

    # Configure
    units = {'day': 'Wh/d', 'hour': 'Wh/h'}[med.cfg['frequency']]
    sample_dict = {'hour': 'H', 'day': 'D', 'week': 'W', 'month': 'M',
                   'year': 'Y'}
    sampling = sample_dict[cfg_plot['frequency']]

    # Define groups to plot
    if result_mng.component_mng.component_name == 'pv':
        groups = {
            # 'irradiance': {
            #     'Irradiance (' + units + '/m2)':
            #     ['global_horizontal_et', 'global_horizontal_surf',
            #      'glob_tilted_surf']},
            'generation': {
                'PV Generation (' + units + ')': ['generation'],
                'Cell Efficiency': ['cell_efficiency']},
            'capacity_factor': {
                'Capacity Factor': ['capacity_factor']}
        }
    elif result_mng.component_mng.component_name in [
            'wind', 'wind-onshore', 'wind-offshore']:
        groups = {
            'generation': {
                'Wind Generation (' + units + ')': ['generation']},
            'capacity_factor': {
                'Capacity Factor': ['capacity_factor']}
        }

    # Get regions
    regions = list(data_src.data.values())[0].indexes['region']

    # Plot for each region
    for reg_label in regions:
        plot_postfix = '_{}_{}'.format(reg_label, cfg_plot['frequency'])

        # Plot per groups
        for group_name, group in groups.items():
            fig = plt.figure()
            ax0 = fig.add_subplot(111)
            any_plot = False
            for k, (label, variable_names) in enumerate(group.items()):
                ax = ax0 if k == 0 else ax0.twinx()
                for iv, variable_name in enumerate(variable_names):
                    da_reg = data_src.get(variable_name)
                    if da_reg is not None:
                        da_reg = da_reg.loc[{'region': reg_label}]
                        da = da_reg.resample(time=sampling).mean(
                            'time', keep_attrs=True)
                        tm = da.indexes['time']
                        ic = (k * len(variable_names) + iv) % len(COLORS)
                        ax.plot(tm, da, label=da.attrs.get('long_name'),
                                color=COLORS[ic])
                        any_plot = True
                ax.set_ylabel(label, fontsize=cfg_plot['fs_default'],
                              color=COLORS[k])
                if len(variable_names) > 1:
                    ax.legend()
            if any_plot:
                ax0.set_xticks(ax.get_xticks()[::2])
                ax0.set_xlabel('time', fontsize=cfg_plot['fs_default'])
                ax0.set_title('{} {} {} {}'.format(
                    reg_label, result_mng.component_mng.name, stage,
                    cfg_plot['frequency']))
                fig_filename = '{}{}{}.{}'.format(
                    group_name, result_postfix, plot_postfix, fig_format)
                fig_filepath = os.path.join(fig_dir, fig_filename)
                if not cfg_plot.get('no_verbose'):
                    med.log.info('Saving figure to {}'.format(fig_filepath))
                fig.savefig(fig_filepath, **cfg_plot['savefigs_kwargs'])

    if cfg_plot['show']:
        plt.show(block=False)


def plot_geo_dist(
        optimizer, lat, lon, capa_pv, capa_wind, mean_penetration=None,
        margin=2., alpha=1., n_markers=4, ms_max=1300., capa_max=10000.,
        trans=0.006, crs=None, facecolor=None, edgecolor=None,
        text=False, units='MW', text_format='{:.1f}', **kwargs):
    """ Plot geographical and technological distribution of RES capacity.

    :param optimizer: Optimizer.
    :param lat: Mean latitude of the regions.
    :param lon: Mean longitude of the regions.
    :param capa_pv: PV capacities.
    :param capa_wind: Wind capacities.
    :param cfg_plot: Plot configuration.
    :param mean_penetration: Penetration rate to add as annotation.
      Default is `None`, in which case no annotation is added.
    :param margin: Margin at the borders of the plot.
      Default is `2` points.
    :param alpha: Alpha value for the markers colors.
      Default is `1`.
    :param n_markers: Number of markers in legend.
      Default is `4`.
    :param ms_max: Maximum marker size.
      Default is `1300`.
    :param capa_max: Maximum capacity in legend.
      Default is `10000`.
    :param trans: Translation factor to seperate
      markers of the different technologies. Default is `0.006`.
    :param crs: Coordinate Reference System. Default is
      :py:class:`cartopy.crs.LambertAzimuthalEqualArea`.
    :param facecolor: Face color of the regions. Default is `'moccasin'`.
    :param edgecolor: Edge color of the regions. Default is `'dimgrey'`.
    :param text: Whether to add text boxes. Default is False.
    :param units: The units of the given data.
    :param text_format: Format of the text boxes.
    :type optimizer: :py:class:`.optimization.OptimizerBase`
    :type lat: sequence
    :type lon: sequence
    :type capa_pv: sequence
    :type capa_wind: sequence
    :type cfg_plot: dict
    :type mean_penetration: float
    :type margin: float
    :type alpha: float
    :type n_markers: int
    :type ms_max: float
    :type capa_max: float
    :type trans: float
    :type crs: :py:class:`cartopy.crs.ABCMeta`
    :type facecolor: str
    :type edgecolor: str
    :type text: bool
    :type units: str
    :type text_format: str
    """
    cfg_plot = load_config(optimizer.med, 'plot')
    facecolor = facecolor or cfg_plot.get('region_facecolor') or 'moccasin'
    edgecolor = edgecolor or cfg_plot.get('region_edgecolor') or 'dimgrey'
    crs = crs or ccrs.LambertAzimuthalEqualArea()

    # Convert geometry to CRS
    geo_src = optimizer.med.geo_src
    gdf = geo_src.get_data(variable_names=GEO_VARIABLE_NAME,
                           **kwargs)[GEO_VARIABLE_NAME]
    gdf_crs = gdf.to_crs(crs.proj4_init)

    fig = plt.figure(figsize=[12, 9])
    ax = plt.axes(projection=crs)
    ax.set_xlabel('Longitude', fontsize=cfg_plot['fs_default'])
    ax.set_ylabel('Latitude', fontsize=cfg_plot['fs_default'])

    # Plot regions
    gdf_crs.plot(ax=ax, facecolor=facecolor, edgecolor=edgecolor)

    # capa_max = np.max([capa_pv.max(), capa_wind.max()])
    leg_size = (np.round(np.linspace(0., 1., n_markers + 1)[1:], 2)
                * np.round(capa_max, -int(np.log10(capa_max)) + 1))
    exp = 1
    fact = ms_max / leg_size.max()**exp
    trans *= (lon.max() - lon.min()) / 2
    trans_lon = trans * (lon.max() - lon.min()) / 2
    trans_lat = trans * (lat.max() - lat.min()) / 2

    # Draw capacity
    s = capa_pv**exp * fact
    ax.scatter(lon - trans * np.sqrt(s), lat, s=s, c=COLORS[0],
               marker='o', alpha=alpha, transform=ccrs.Geodetic())

    s = capa_wind**exp * fact
    ax.scatter(lon + trans * np.sqrt(s), lat, s=s, c=COLORS[1],
               marker='o', alpha=alpha, transform=ccrs.Geodetic())

    # Add text
    if text:
        for k, (c_pv, c_wind) in enumerate(zip(capa_pv, capa_wind)):
            transform = ccrs.Geodetic()._as_mpl_transform(ax)
            # Annotate PV
            t_lon = lon[k] - 30 * trans_lon
            t_lat = lat[k] - 1.5 * trans_lat
            ax.annotate(text_format.format(c_pv.values), xy=(t_lon, t_lat),
                        xycoords=transform, fontsize='xx-large')
            # Annotate wind
            t_lon = lon[k] + 15 * trans_lon
            t_lat = lat[k] - 1.5 * trans_lat
            ax.annotate(text_format.format(c_wind.values), xy=(t_lon, t_lat),
                        xycoords=transform, fontsize='xx-large')

    # Draw legend
    h_pv = [plt.plot([-1e9], [-1e9], linestyle='',
                     marker='$\mathrm{PV}$', color='k', markersize=20)[0]]
    h_wind = [plt.plot([-1e9], [-1e9], linestyle='',
                       marker='$\mathrm{Wind}$', color='k', markersize=40)[0]]
    h_pv += [plt.scatter(
        [], [], s=leg_size[s]**exp * fact, c=COLORS[0],
        marker='o', alpha=alpha) for s in np.arange(n_markers)]
    h_wind += [plt.scatter([], [], s=leg_size[s]**exp * fact,
                           c=COLORS[1], marker='o',
                           alpha=alpha) for s in np.arange(n_markers)]
    l_pv, l_wind = [''], ['']
    l_pv += ['' for s in np.arange(n_markers)]
    l_wind += ['{:} MW'.format(int(leg_size[s])) for s in np.arange(n_markers)]
    l_wind += ['{:} {}'.format(int(leg_size[s]), units)
               for s in np.arange(n_markers)]
    ax.legend(h_pv + h_wind, l_pv + l_wind, loc='best', handletextpad=1.,
              labelspacing=2.5, borderpad=1.5, ncol=2, columnspacing=1.5)

    if mean_penetration is not None:
        starget = '$\mu^* = {:.1f}\%$'.format(mean_penetration * 100)
        plt.annotate(starget, xy=(0.7, 0.02), xycoords='axes fraction',
                     fontsize=cfg_plot['fs_latex'])

    # Set extent
    tb = gdf_crs.total_bounds
    extent = tb[0], tb[2], tb[1], tb[3]
    ax.set_extent(extent, crs)

    return fig


def plot_band_spectrum(med, ds, filt, time_slice=None,
                       plot_freq=['Y', 'D', 'H'], var_ylims={},
                       add_legend=False, **kwargs):
    """Plot band spectrum, i.e. the integration of power spectrum over
    frequency bands.

    :param med: Mediator.
    :param ds: Dataset of output variables of components.
    :param filt: Filtered data.
    :param time_slice: Period to select.
    :param plot_freq: Frequencies for which to plot.
      Default is `['Y', 'D', 'H']`.
    :param var_ylims: Mapping of y-axis limits to result-manager names.
    :param add_legend: Whether to add the legend. Default is `False`.
    :type med: :py:class:`.mediator.Mediator`
    :type ds: mapping
    :type filt: mapping
    :type time_slice: slice
    :type plot_freq: sequence
    :type var_ylims: mapping
    :type add_legend: bool
    """
    cfg_plot = load_config(med, 'plot')
    fig_format = cfg_plot['savefigs_kwargs']['format']
    lw = {'Y': 3, 'D': 2, 'H': 1}
    zo = {'Y': 3, 'D': 2, 'H': 1}
    labels = {'Y': 'Yearly-mean', 'D': 'Daily-mean', 'H': 'Hourly'}
    var_labels = {'demand': '', 'capacity_factor': 'Capacity Factor'}
    comp_labels = {
        'demand': 'Demand', 'pv': 'PV', 'wind': 'Wind',
        'wind-onshore': 'Onshore Wind', 'wind-offshore': 'Offshore Wind'}
    figsize = kwargs.get('figsize') or rcParams['figure.figsize']
    for component_mng_name, ds_comp in ds.items():
        component_mng = med[component_mng_name]
        fig_dir = med.cfg.get_plot_directory(component_mng, **kwargs)
        for res_variable_name, da in ds_comp.items():
            result_mng = component_mng[res_variable_name]
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)
            filt_var = filt[component_mng_name][res_variable_name]
            for ifr, freq in enumerate(filt_var):
                dum = filt_var[freq]
                if time_slice is not None:
                    dum = dum.sel(time=time_slice)
                if ifr == 0:
                    ts = dum
                else:
                    ts += dum
                if freq in plot_freq:
                    ax.plot(ts.time, ts, linewidth=lw[freq], zorder=zo[freq],
                            label=labels[freq])

            ylabel = '{} {}'.format(
                comp_labels[component_mng_name], var_labels[res_variable_name])
            units = da.attrs.get('units')
            if units:
                ylabel += ' ({})'.format(units)
            elif res_variable_name == 'capacity_factor':
                ylabel += ' (%)'
            ax.set_ylabel(ylabel, fontsize=cfg_plot['fs_default'])
            # Add limits
            ylim = var_ylims.get(res_variable_name)
            if ylim is not None:
                ax.set_ylim(ylim)

            # Add legend only if regions not present
            if add_legend:
                ax.legend(loc='upper right')

            # Save figure
            result_postfix = result_mng.get_data_postfix(**kwargs)
            if time_slice is not None:
                result_postfix += '_{}_{}'.format(
                    time_slice.start, time_slice.stop)
            fig_filename = 'filtered_{}{}.{}'.format(
                result_mng.name, result_postfix, fig_format)
            fig_filepath = os.path.join(fig_dir, fig_filename)
            if not cfg_plot.get('no_verbose'):
                med.log.info('Saving figure to {}'.format(fig_filepath))
            fig.savefig(fig_filepath, **cfg_plot['savefigs_kwargs'])

    if cfg_plot['show']:
        plt.show(block=False)
